import json
import traceback
from typing import Tuple
from papys.actions.core import PAction
from papys.request_response import Request, Response


class StaticJsonAction(PAction):
    """
    Class providing a simple JSON response.
    You can use this class to transfer a JSON object, which is then written to the response.
    You can use it for test purposes, but also productively if you want to deliver a short static content for an endpoint.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        json (dict) : The JSON to deliver as response content.
        status_codes (dict): You can customize the returned status in case of success and an error. Default: { 'success': 200, 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        json: dict = {},
        status_codes={"success": 200, "error": 500},
    ):
        self._json = json
        self._status_codes = status_codes
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            resp.to_convert = self._json
            resp.json = json.dumps(self._json)
            resp.status_code = self._status_codes["success"]
            return self._status_codes["success"], req, resp
        except Exception as err:
            req.logger.log_error(
                "StaticJsonAction could not processed.", traceback.format_exc(), 121, req
            )
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp


# TODO
# class StaticFileAction(PAction)
# return the File from path. Basedirectory is set in the global config.


from urllib.parse import urlencode
import papys.core
from papys.route import PRoute as R
from papys.actions.flows import RedirectAction
from papys.http_methods import GET
from papys.hooks import PHook
from papys.security_hooks import KcOidcAcfRouteGuardHook, KcOidcCcfRouteGuardHook
from papys.actions.authentication import KcOidcAcfCallbackAction, KcOidcAcfLogoutAction


class KcOidcAcfFactory:
    """
    A class providing Keycloak OIDC security on the Papys app
    with Authorization Code Flow.
    Use this factory with the with statement:

    Minimum usage:

    with KcOidcAcfFactory() as oidc:
        oidc.server_host = "https://host-of-papys-rest-api.com"
        oidc.auth_url = "https://host-of-keycloak.com/auth/realms/"
        oidc.client_id = "client-id"
        oidc.client_secret = "the-secret-from-keycloak"
        oidc.realm = "your-realm"
        oidc.app_redirect_url = "http://example.com/where-your-app-is"

    Full configuration:

    with KcOidcAcfFactory() as oidc:
        oidc.server_host = "https://host-of-papys-rest-api.com"
        oidc.callback_path = "/callback"
        oidc.login_path = "/login"
        oidc.logout_path = "/logout"
        oidc.auth_url = "https://host-of-keycloak.com/auth/realms/"
        oidc.client_id = "client-id"
        oidc.client_secret = "the-secret-from-keycloak"
        oidc.realm = "your-realm"
        oidc.redirect_to_login = True
        oidc.app_redirect_url = "http://example.com/where-your-app-is"

    Get your guard hook:

    oicd_guard = oidc.get_route_guard_hook()

    Now you can use this hook in your route(s) you want to protect:

    protected_route = R("/protected", oicd_guard) >> [
        (GET, YourActionHere("My action"))
    ]

    """

    def __init__(self):
        self._server_host: str | None = None
        self._callback_path: str | None = "/callback"
        self._login_path: str | None = "/login"
        self._logout_path: str | None = "/logout"
        self._client_id: str | None = None
        self._client_secret: str | None = None
        self._auth_url: str | None = None
        self._realm: str | None = None
        self._app_redirect_url: str | None = None
        self._redirect_to_login: bool = False

        self._auth_url_postfix: str | None = "protocol/openid-connect/auth"
        self._is_open: bool = True

    @property
    def server_host(self):
        return self._server_host

    @server_host.setter
    def server_host(self, value: str | None):
        self._server_host = value

    @property
    def callback_path(self):
        return self._callback_path

    @callback_path.setter
    def callback_path(self, value: str | None):
        self._callback_path = value

    @property
    def login_path(self):
        return self._login_path

    @login_path.setter
    def login_path(self, value: str | None):
        self._login_path = value

    @property
    def logout_path(self):
        return self._logout_path

    @logout_path.setter
    def logout_path(self, value: str | None):
        self._logout_path = value

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value: str | None):
        self._client_id = value

    @property
    def client_secret(self):
        return self._client_secret

    @client_secret.setter
    def client_secret(self, value: str | None):
        self._client_secret = value

    @property
    def auth_url(self):
        return self._auth_url

    @auth_url.setter
    def auth_url(self, value: str | None):
        self._auth_url = value

    @property
    def realm(self):
        return self._realm

    @realm.setter
    def realm(self, value: str | None):
        self._realm = value

    @property
    def app_redirect_url(self):
        return self._app_redirect_url

    @app_redirect_url.setter
    def app_redirect_url(self, value: str | None):
        self._app_redirect_url = value

    @property
    def redirect_to_login(self):
        return self._redirect_to_login

    @redirect_to_login.setter
    def redirect_to_login(self, value: bool):
        self._redirect_to_login = value

    def get_route_guard_hook(self) -> PHook:
        config = {
            "base_url": f"{self.auth_url}/{self._realm}/",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_to_login": self.redirect_to_login,
            "login_path": self.login_path,
        }
        return KcOidcAcfRouteGuardHook(config)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        redirect_url = f"{self.server_host + self.callback_path}"
        if self._is_open:
            query_param = {
                "client_id": self.client_id,
                "response_type": "code",
                "scope": "openid",
                "redirect_uri": redirect_url,
            }
            login_route = R(self.login_path) >> [
                (
                    GET,
                    RedirectAction(
                        "Redirect to IAM login url",
                        url=f"{self.auth_url}/{self._realm}/{self._auth_url_postfix}?{urlencode(query_param, safe=':/')}",
                    ),
                )
            ]
            papys.core.add_route(login_route)

            config = {
                "base_url": f"{self.auth_url}/{self._realm}/",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "redirect_url": redirect_url,
                "app_redirect_url": self.app_redirect_url,
            }

            callback_route = R(self.callback_path) >> [
                (GET, KcOidcAcfCallbackAction("Handle IAM callback", config))
            ]
            papys.core.add_route(callback_route)

            logout_config = {
                "base_url": f"{self.auth_url}/{self._realm}/",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }

            logout_route = R(self.logout_path) >> [
                (GET, KcOidcAcfLogoutAction("Logout user at IAM", logout_config))
            ]
            papys.core.add_route(logout_route)

            self._is_open = False


class KcOidcCcfFactory:
    """
    A class providing Keycloak OIDC security on the Papys app
    with Client Credential Flow (M2M).
    Use this factory with the with statement:

    Usage:

    with KcOidcCcfFactory() as oidc:
        oidc.auth_url = "https://host-of-keycloak.com/auth/realms/"
        oidc.realm = "your-realm"

    Get your guard hook:

    oicd_guard = oidc.get_route_guard_hook()

    Now you can use this hook in your route(s) you want to protect:

    protected_route = R("/protected", oicd_guard) >> [
        (GET, YourActionHere("My action"))
    ]

    """

    def __init__(self):
        self._auth_url: str | None = None
        self._realm: str | None = None
        self._client_id: str | None = None
        self._client_secret: str | None = None

        self._auth_url_postfix: str | None = "protocol/openid-connect/token/introspect"
        self._is_open: bool = True

    @property
    def auth_url(self):
        return self._auth_url

    @auth_url.setter
    def auth_url(self, value: str | None):
        self._auth_url = value

    @property
    def realm(self):
        return self._realm

    @realm.setter
    def realm(self, value: str | None):
        self._realm = value

    @property
    def client_id(self):
        return self._client_id
    
    @client_id.setter
    def client_id(self, value: str):
        self._client_id = value

    @property
    def client_secret(self):
        return self._client_secret
    
    @client_secret.setter
    def client_secret(self, value: str):
        self._client_secret = value


    def get_route_guard_hook(self) -> PHook:
        config = {
            "validate_url": f"{self.auth_url}/{self._realm}/{self._auth_url_postfix}",
            "client_id": self.client_id,
            "client_secret": self.client_secret
        }
        return KcOidcCcfRouteGuardHook(config)

    def __enter__(self):
        return self

    def __exit__(self, *args):        
        if self._is_open:
            self._is_open = False



import traceback
import requests
from typing import Tuple
from collections.abc import Iterable
from papys.http_methods import GET, POST, PUT, DELETE
from papys.hooks import PHook
from papys.request_response import (
    Request,
    Response,
    COOKIE_SAME_SITE_STRICT,
)

AUTH_ALLOW = "allow"
AUTH_DENY = "deny"


class KcOidcAcfRouteGuardHook(PHook):
    """
    PHook implementation to secure a route with Keycloak OICD.
    This hook will be generated by the KcOIDCFactory class.
    It's recommended to not instance this class yourself.

    Args:
        config (dict): The OICD configuration for checking the authoritiy.
        status_codes (dict): {'success': 200, 'unauthorized': 401, 'error': 500} is default. You can provide other codes to return in case of success or an exception.
    """

    def __init__(
        self,
        config=None,
        status_codes: dict = {"success": 200, "unauthorized": 401, "error": 500},
    ):
        self._config = config
        self._status_codes = status_codes
        super().__init__()

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        try:
            access_token = req.parsed_cookie.get("access_token", None)
            if access_token:
                user_info_response = requests.get(
                    self._config.get("base_url") + "/protocol/openid-connect/userinfo",
                    headers={"Authorization": f"Bearer {access_token}"},
                )
                if user_info_response.status_code == 200:
                    req.user_info = user_info_response.json()
                    return True, self._status_codes["success"], req, resp
                else:
                    payload = {
                        "grant_type": "refresh_token",
                        "client_id": self._config.get("client_id"),
                        "client_secret": self._config.get("client_secret"),
                        "refresh_token": req.parsed_cookie.get("refresh_token", None),
                    }

                    response = requests.post(
                        self._config.get("base_url") + "/protocol/openid-connect/token",
                        data=payload,
                    )
                    response_data = response.json()

                    if response.status_code == 200:
                        resp.add_cookie(
                            "access_token",
                            response_data["access_token"],
                            resp.create_cookie_attributes(
                                http_only=True,
                                same_site=COOKIE_SAME_SITE_STRICT,
                                secure=True,
                            ),
                        )

                        resp.add_cookie(
                            "refresh_token",
                            response_data["refresh_token"],
                            resp.create_cookie_attributes(
                                http_only=True,
                                same_site=COOKIE_SAME_SITE_STRICT,
                                secure=True,
                            ),
                        )
                        user_info_response = requests.get(
                            self._config.get("base_url")
                            + "/protocol/openid-connect/userinfo",
                            headers={
                                "Authorization": f"Bearer {response_data['access_token']}"
                            },
                        )
                        if user_info_response.status_code == 200:
                            req.user_info = user_info_response.json()
                            return True, self._status_codes["success"], req, resp
                        else:
                            req.logger.log_warning(
                                f"Blocked unauthorized access, 401. Token refresh successful but get user info failes. access_token: {response_data['access_token']}",
                                131,
                                req,
                            )
                            return False, self._status_codes["unauthorized"], req, resp
                    else:
                        req.logger.log_warning(
                            f"Blocked unauthorized access, 401. Token refresh failed. access_token: {access_token}",
                            131,
                            req,
                        )
                        if self._config.get("redirect_to_login", False):
                            resp.set_header("Location", self._config.get("login_path"))
                            return False, 302, req, resp
                        else:
                            return False, self._status_codes["unauthorized"], req, resp
            else:
                req.logger.log_warning(
                    "Blocked unauthorized access, 401. No access_token found.", 132, req
                )
                if self._config.get("redirect_to_login", False):
                    resp.set_header("Location", self._config.get("login_path"))
                    return False, 302, req, resp
                else:
                    return False, self._status_codes["unauthorized"], req, resp
        except Exception as err:
            req.logger.log_error(
                "KcOidcAcfRouteGuardHook could not processed.",
                traceback.format_exc(),
                133,
                req,
            )
            return False, self._status_codes["error"], req, resp


class KcOidcCcfRouteGuardHook(PHook):
    """
    PHook implementation to secure a route with Keycloak OICD.
    This hook will be generated by the KcOIDCFactory class.
    It's recommended to not instance this class yourself.

    Args:
        config (dict): The OICD configuration for checking the authoritiy.
        status_codes (dict): {'success': 200, 'unauthorized': 401, 'error': 500} is default. You can provide other codes to return in case of success or an exception.
    """

    def __init__(
        self,
        config=None,
        status_codes: dict = {"success": 200, "unauthorized": 401, "error": 500},
    ):
        self._config = config
        self._status_codes = status_codes
        super().__init__()

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        try:
            access_token = req.get_bearer_token()
            if access_token:
                data = {
                    "token": access_token,
                    "client_id": self._config.get("client_id", ""),
                    "client_secret": self._config.get("client_secret", ""),
                }
                token_response = requests.post(
                    self._config.get("validate_url"),
                    data=data,
                )
                token_valid_response = token_response.json()
                if token_valid_response.get("active"):
                    req.user_info = token_valid_response
                    return True, self._status_codes["success"], req, resp
                else:
                    req.logger.log_warning(
                        "Blocked unauthorized access, 401. Invalid access token",
                        132,
                        req,
                    )
                    return False, self._status_codes["unauthorized"], req, resp
            else:
                req.logger.log_warning(
                    "Blocked unauthorized access, 401. No access_token found.", 133, req
                )
                return False, self._status_codes["unauthorized"], req, resp

        except Exception as err:
            req.logger.log_error(
                "KcOidcCcfRouteGuardHook could not processed.",
                traceback.format_exc(),
                134,
                req,
            )
            return False, self._status_codes["error"], req, resp


class KcOidcAcfAuthorizationHook(PHook):
    def __init__(self):
        super().__init__()
        self._type: str = AUTH_ALLOW
        self._user_info_group_attribute_name: str = "roles"
        self._GET_groups: set = set()
        self._POST_groups: set = set()
        self._PUT_groups: set = set()
        self._DELETE_groups: set = set()
        self._add_user_sub_to_body: bool = False
        self._user_id_body_attribute_name: str | None = "user_id"

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, value: str) -> None:
        self._type = value

    @property
    def user_info_group_attribute_name(self) -> str:
        return self._user_info_group_attribute_name

    @user_info_group_attribute_name.setter
    def user_info_group_attribute_name(self, value: str) -> None:
        self._user_info_group_attribute_name = value

    @property
    def GET_groups(self) -> set:
        return self._GET_groups

    @GET_groups.setter
    def GET_groups(self, value: set) -> None:
        self._GET_groups = value

    @property
    def POST_groups(self) -> set:
        return self._POST_groups

    @POST_groups.setter
    def POST_groups(self, value: set) -> None:
        self._POST_groups = value

    @property
    def PUT_groups(self) -> set:
        return self._PUT_groups

    @PUT_groups.setter
    def PUT_groups(self, value: set) -> None:
        self._PUT_groups = value

    @property
    def DELETE_groups(self) -> set:
        return self._DELETE_groups

    @DELETE_groups.setter
    def DELETE_groups(self, value: set) -> None:
        self._DELETE_groups = value

    @property
    def add_user_sub_to_body(self) -> bool:
        return self._add_user_sub_to_body

    @add_user_sub_to_body.setter
    def add_user_sub_to_body(self, value: bool) -> None:
        self._add_user_sub_to_body = value

    @property
    def user_id_body_attribute_name(self) -> str | None:
        return self._user_id_body_attribute_name

    @user_id_body_attribute_name.setter
    def user_id_body_attribute_name(self, value: str | None) -> None:
        self._user_id_body_attribute_name = value

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        try:
            user_groups = req.user_info[self.user_info_group_attribute_name]
            if isinstance(user_groups, Iterable):
                user_groups_set = set(user_groups)
                match req.request_method:
                    case "GET":
                        group_check = len(user_groups_set & self.GET_groups)
                    case "POST":
                        group_check = len(user_groups_set & self.POST_groups)
                    case "PUT":
                        group_check = len(user_groups_set & self.PUT_groups)
                    case "DELETE":
                        group_check = len(user_groups_set & self.DELETE_groups)
                    case _:
                        raise ValueError(
                            f"Not supported http method found: {req.request_method}."
                        )

                match (self.type, group_check):
                    case ("allow", n) if n > 0:
                        if self.add_user_sub_to_body and req.body_json:
                            req.body_json[self.user_id_body_attribute_name] = (
                                req.user_info["sub"]
                            )
                        return True, 200, req, resp
                    case ("allow", n) if n < 1:
                        return False, 401, req, resp
                    case ("deny", n) if n < 1:
                        if self.add_user_sub_to_body and req.body_json:
                            req.body_json[self.user_id_body_attribute_name] = (
                                req.user_info["sub"]
                            )
                        return True, 200, req, resp
                    case ("deny", n) if n > 0:
                        return False, 401, req, resp
                    case _:
                        return False, 401, req, resp
            else:
                return False, 401, req, resp
        except Exception as err:
            req.logger.log_error(
                "Failed to check user groups.", traceback.format_exc(), 136, req
            )
            return False, 500, req, resp


class KcAcfCcfSwitchHook(PHook):
    def __init__(self, *, acfHook: PHook, ccfHook: PHook):
        super().__init__()
        self._acfHook: PHook = acfHook
        self._ccfHook: PHook = ccfHook

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        try:
            if req.get_bearer_token():
                ccf_result = self._ccfHook.process_hook(req, resp)
                if ccf_result[0]:
                    req.authentication_method = "kcCcf"
                    return ccf_result
            acf_result = self._acfHook.process_hook(req, resp)
            if acf_result[0]:
                req.authentication_method = "kcAcf"
            return acf_result
        except Exception as err:
            req.logger.log_error(
                "Failed to process KcAcfCcfSwitchHook.", traceback.format_exc(), 137, req
            )
            return False, 500, req, resp



import traceback
from typing import Tuple
from papys.actions.core import PAction
from papys.request_response import Request, Response


class RedirectAction(PAction):
    """
    Class providing a client redirect.
    This implementation sets the http header "Location" to the desired URL
    and the http status code to the selected status code. E.g. 302.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        url (str): The url for the http header 'Location' for the redirect.
        redirect_status_code (int): The desired http status code for the redirection. Default: 302.
        status_codes (dict): You can customize the returned status in case of an error. Default: { 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        url: str = "/",
        redirect_status_code=302,
        status_codes={"error": 500},
    ):
        self._url = url
        self.redirect_status_code = redirect_status_code
        self._status_codes = status_codes
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            resp.set_header("Location", self._url)
            resp.status_code = self.redirect_status_code
            return self.redirect_status_code, req, resp
        except Exception as err:
            req.logger.log_error(
                "RedirectAction could not processed.", traceback.format_exc(), 122, req
            )
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp



import json
import jsonschema
import traceback
from typing import Tuple
from jsonschema import validate
from papys.actions.core import PAction
from papys.request_response import Request, Response


class JsonRespValidatorAction(PAction):
    """
    Class providing a default JSON schema validation.
    This implementation uses the JSON schema validation of the standard library to check a JSON body against a JSON schema.
    It is recommended that both incoming and outgoing content is validated.
    This ensures that no unwanted information is sent out and minimizes the risk of incorrect or dangerous information being processed.
    This implementation checks the content of Response.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        schema (dict): The JSON schema to validate against. See more: https://json-schema.org/
        status_codes (dict): You can customize the returned status in case of success and an error. Default: { 'success': 200, 'error': 500 }.
        return_full_error (bool): If True the whole error message is returned in cas of an validation error. Use True only for debugging. There is a risk of sensitive information being sent to the client unintentionally. The default is False.
    """

    def __init__(
        self,
        name: str = "",
        schema: dict = {},
        status_codes={"success": 200, "error": 500},
        return_full_error: bool = False,
    ):
        self._schema = schema
        self._status_codes = status_codes
        self._return_full_error = return_full_error
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            content = (
                resp.to_convert
                if resp.to_convert is not None
                else json.loads(resp.json)
            )
            validate(instance=content, schema=self._schema)
            resp.status_code = self._status_codes["success"]
            return self._status_codes["success"], req, resp
        except jsonschema.exceptions.ValidationError as format_error:
            resp.is_error = True
            resp.error = (
                format_error if self._return_full_error else format_error.message
            )
            resp.status_code = self._status_codes["error"]
            return self._status_codes["error"], req, resp
        except Exception as err:
            req.logger.log_error(
                "JsonRespValidatorAction could not processed.", traceback.format_exc(), 123, req
            )
            resp.is_error = True
            resp.error = err
            resp.status_code = self._status_codes["error"]
            return self._status_codes["error"], req, resp


class JsonReqValidatorAction(PAction):
    """
    Class providing a default JSON schema validation.
    This implementation uses the JSON schema validation of the standard library to check a JSON body against a JSON schema.
    It is recommended that both incoming and outgoing content is validated.
    This ensures that no unwanted information is sent out and minimizes the risk of incorrect or dangerous information being processed.
    This implementation checks the content of Request.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        schema (dict) : The JSON schema to validate against. See more: https://json-schema.org/
        status_codes (dict): You can customize the returned status in case of success and an error. Default: { 'success': 200, 'error': 500 }.
        return_full_error (bool): If True the whole error message is returned in cas of an validation error. Default is False.
    """

    def __init__(
        self,
        name: str = "",
        schema: dict = {},
        status_codes={"success": 200, "error": 500},
        return_full_error: bool = False,
    ):
        self._schema = schema
        self._status_codes = status_codes
        self._return_full_error = return_full_error
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            content = req.body_json
            validate(instance=content, schema=self._schema)
            resp.status_code = self._status_codes["success"]
            return self._status_codes["success"], req, resp
        except jsonschema.exceptions.ValidationError as format_error:
            resp.is_error = True
            resp.error = (
                format_error if self._return_full_error else format_error.message
            )
            resp.status_code = self._status_codes["error"]
            return self._status_codes["error"], req, resp
        except Exception as err:
            req.logger.log_error(
                "JsonReqValidatorAction could not processed.", traceback.format_exc(), 124, req
            )
            resp.is_error = True
            resp.error = err
            resp.status_code = self._status_codes["error"]
            return self._status_codes["error"], req, resp



from papys.hooks import PHook


class PRoute:
    """
    A route define a path, maybe holds sub-routes and the given actions mapped to the http method.
    Chain routes with: |. Route('/path') | Route('/sub-path') gets '/path/sub-path'.
    Add the action with >>. Example:
    Route('/path') >> [ ( GET, Action ) ]

    Args:
        path (str): String with the relativ path. Best practise: Start with / and end without. This doesn't get confusing when chaining routes Example. You can use variables with brackets. Example: /user/{user_id}/document/{doc_id}. You can use regular expression too. Variables are accessible with (example) Request.path_variables.get('user_id')
        hook (PHook): Optional a hook which will be processed befor the (sub-) route action will be processed. You cann chain hooks as well with: hook1 >> hook2.
    """

    def __init__(self, path: str = "/", hook: PHook | None = None):
        self._path = path
        self._sub_routes = []
        self._actions = {}
        self._hook = hook

    @property
    def path(self):
        return self._path

    @property
    def sub_routes(self):
        return self._sub_routes

    @property
    def actions(self):
        return self._actions

    @actions.setter
    def actions(self, value):
        self._actions = value

    @property
    def hook(self) -> PHook:
        return self._hook

    # |
    def __or__(self, other):
        if isinstance(other, PRoute):
            self.sub_routes.append(other)
        else:
            raise TypeError("The operand must be an instance of PRoute.")
        return self

    # >>
    def __rshift__(self, other):
        if isinstance(other, list):
            for element in other:
                if not isinstance(element, tuple):
                    raise TypeError(
                        "The operand must be an instance of list with tuple."
                    )
            self.actions = other
            # TODO Expand with method and action check?
        else:
            raise TypeError("The operand must be an instance of list.")
        return self



from wsgiref.simple_server import make_server
import papys.core as core


def run(app=core.app, host="localhost", port=8000):
    """
    IMPORTANT!
    Please you this function just in case of testing. You a production ready WSGI server for running the application in production.

    Args:
        app (function): The WSGI compatible function. Default: core.app
        host (str): The host of the server. Default: 'localhost'.
        port (int): The port of the server. Default: 8000.
    """
    httpd = make_server(host, port, app)
    print(f"Server is running on: {host}:{port}")
    print(f"WARNING: Don't use in production environement!")
    print(f"Use a production ready WSGI server instead.")
    httpd.serve_forever()



class PConfig:
    """
    Class to configure the Papys application.

    Args:
        log_level (int): 1: Errors only. 2: Errors and warnings. 3: All incl. information. Default is 2.
        post_convert_201 (bool): True: If status is 200 and http method is post, then http status is set to 201. Default is True.
        return_error500_body (bool): True: In case of status 500 a return body from Response.error_message field is created. Default is False.
        accept_default_lang (str): If the http accept lang header is missing this one is set. Default is 'US-en'.

    """

    def __init__(
        self,
        log_level: int = 2,
        post_convert_201: bool = True,
        return_error500_body: bool = False,
        accept_default_lang: str = "en-US",
    ):
        self._log_level = log_level
        self._post_convert_201 = post_convert_201
        self._return_error500_body = return_error500_body
        self._accept_default_lang = accept_default_lang

    @property
    def log_level(self) -> int:
        return self._log_level

    @log_level.setter
    def log_level(self, value: int):
        self._log_level = value

    @property
    def post_convert_201(self) -> bool:
        return self._post_convert_201

    @post_convert_201.setter
    def post_convert_201(self, value: bool):
        self._post_convert_201 = value

    @property
    def return_error500_body(self) -> bool:
        return self._return_error500_body

    @return_error500_body.setter
    def return_error500_body(self, value: bool):
        self._return_error500_body = value

    @property
    def accept_default_lang(self) -> str:
        return self._accept_default_lang

    @accept_default_lang.setter
    def accept_default_lang(self, value: str):
        self._accept_default_lang = value



import json
import re
import urllib.parse
import traceback
from typing import Tuple, Any
from papys.route import PRoute
from papys.actions.core import PAction
from papys.http_methods import GET, POST, PUT, DELETE
from papys.request_response import Request, Response
from papys.hooks import PHook
from papys.config import PConfig
from papys.utils import PLogger, PathCache


_http_code_mapping = {
    200: "200 OK",
    201: "201 Created",
    202: "202 Accepted",
    203: "203 Non-Authoritative",
    204: "204 No Content",
    205: "205 Reset Content",
    206: "206 Partial Content",
    300: "300 Multiple Choices",
    301: "301 Moved Permanently",
    302: "302 Found (Moved Temporarily)",
    303: "303 See Other",
    304: "304 Not Modified",
    305: "305 Use Proxy",
    307: "307 Temporary Redirect",
    308: "308 Permanent Redirect",
    400: "400 Bad Request",
    401: "401 Unauthorized",
    402: "402 Payment Required",
    403: "403 Forbidden",
    404: "404 Not Found",
    405: "405 Method Not Allowed",
    406: "406 Not Acceptable",
    407: "407 Proxy Authentication Required",
    408: "408 Request Timeout",
    409: "409 Conflict",
    410: "410 Gone",
    411: "411 Length Required",
    412: "412 Precondition Failed",
    413: "413 Payload Too Large",
    414: "414 URI Too Long",
    415: "415 Unsupported Media Type",
    416: "416 Range Not Satisfiable",
    417: "417 Expectation Failed",
    421: "421 Misdirected Request",
    422: "422 Unprocessable Entity",
    423: "423 Locked",
    424: "424 Failed Dependency",
    425: "425 Too Early",
    426: "426 Upgrade Required",
    428: "428 Precondition Required",
    429: "429 Too Many Requests",
    431: "431 Request Header Fields Too Large",
    451: "451 Unavailable For Legal Reasons",
    500: "500 Internal Server Error",
    501: "501 Not Implemented",
    502: "502 Bad Gateway",
    503: "503 Service Unavailable",
    504: "504 Gateway Timeout",
    505: "505 HTTP Version not supported",
    506: "506 Variant Also Negotiates",
    507: "507 Insufficient Storage",
    508: "508 Loop Detected",
    509: "509 Bandwidth Limit Exceeded",
    510: "510 Not Extended",
    511: "511 Network Authentication Required",
}

_get_actions = {}
_post_actions = {}
_put_actions = {}
_delete_actions = {}

_config = PConfig()
_logger = PLogger()
_path_cache = PathCache()

_initialize_hook = None
_finalize_hook = None


def _compile_path(path: str):
    pattern = re.sub(r"\{(\w+)\}", r"(?P<\1>[^/]+)", path)
    return re.compile(pattern)


def _match_path(look_up: dict, path: str) -> Tuple[bool, PAction, Any, PHook]:
    assert look_up is not None, "Dictionary should not be None."
    for key, value in look_up.items():
        match = key.fullmatch(path)
        if match:
            return True, value[0], match.groupdict(), value[1]

    return False, None, None, None


def _add_actions(path: str, actions: list, hook: PHook = None):
    if actions is not None:
        for action in actions:
            match action[0]:
                case "GET":
                    _get_actions[_compile_path(path)] = action[1], hook
                case "POST":
                    _post_actions[_compile_path(path)] = action[1], hook
                case "PUT":
                    _put_actions[_compile_path(path)] = action[1], hook
                case "DELETE":
                    _delete_actions[_compile_path(path)] = action[1], hook


def _add_action(path: str, action: PAction, hook: PHook = None):
    if action is not None:
        match action[0]:
            case "GET":
                _get_actions[_compile_path(path)] = action[1], hook
            case "POST":
                _post_actions[_compile_path(path)] = action[1], hook
            case "PUT":
                _put_actions[_compile_path(path)] = action[1], hook
            case "DELETE":
                _delete_actions[_compile_path(path)] = action[1], hook


def _add_sub_routes(path: str, routes: list, hook: PHook):
    if routes is not None:
        for route_entry in routes:
            sub_hook = (
                hook >> route_entry.hook if hook is not None else route_entry.hook
            )
            for action_entry in route_entry.actions:
                _add_action(path + route_entry.path, action_entry, sub_hook)
            if route_entry.sub_routes:
                _add_sub_routes(
                    path + route_entry.path, route_entry.sub_routes, sub_hook
                )


def _execute_sub_actions(
    statusCode: int, req: Request, resp: Response, sub_actions: list
) -> Tuple[int, Request, Response]:
    result = statusCode, req, resp
    for sub_action in sub_actions:
        if sub_action[0] == statusCode:
            result = sub_action[1].process(result[1], result[2])
            match sub_action[1].sub_actions:
                case list():
                    result = _execute_sub_actions(
                        result[0], result[1], result[2], sub_action[1].sub_actions
                    )
                case _ if sub_action[1].sub_actions is not None:
                    result = _execute_sub_actions(
                        result[0], result[1], result[2], [sub_action[1].sub_actions]
                    )
    return result


def _execute_action_path(req: Request, resp: Response, action: PAction):
    result = action.process(req, resp)
    match action.sub_actions:
        case list():
            result = _execute_sub_actions(
                result[0], result[1], result[2], action.sub_actions
            )
        case PAction():
            result = _execute_sub_actions(
                result[0], result[1], result[2], [action.sub_actions]
            )
    return result


def _execute_hook(
    hook: PHook, req: Request, resp: Response
) -> Tuple[bool, int, Request, Response]:
    hook_result = hook.process_hook(req, resp)
    if not hook_result[0]:
        return (False,) + hook_result[1:]

    if hook.sub_hook is not None:
        sub_hook_result = _execute_hook(hook.sub_hook, hook_result[2], hook_result[3])
        if not sub_hook_result[0]:
            return (False,) + sub_hook_result[1:]
        else:
            return (True,) + sub_hook_result[1:]
    else:
        return (True,) + hook_result[1:]


def _handle_request(req: Request, resp: Response, actions: dict, alt: list = []):
    cache_result = _path_cache.check(req.request_method, req.path)
    if cache_result[0]:
        action_info = cache_result
    else:
        action_info = _match_path(actions, req.path)

    if not action_info[0]:
        resp.status_code = 404
        for alt_action in alt:
            check_result = _match_path(alt_action, req.path)
            if check_result[0]:
                resp.status_code = 405
        return resp.status_code, req, resp

    if cache_result[0] == False:
        _path_cache.add(req.request_method, req.path, action_info[1:])

    action = action_info[1]
    req.path_variables = action_info[2]
    hook = action_info[3]
    if hook is not None:
        hook_result = _execute_hook(hook, req, resp)
        if hook_result[0] is True:
            result = _execute_action_path(hook_result[2], hook_result[3], action)
        else:
            return hook_result[1:]
    else:
        result = _execute_action_path(req, resp, action)
    return result


def _create_request(environ: dict) -> Request:
    req = Request()
    req.logger = _logger
    req.global_config = _config
    req.path = environ.get("PATH_INFO", "")
    req.server_name = environ.get("SERVER_NAME", "")
    req.gateway_interface = environ.get("GATEWAY_INTERFACE", "")
    req.server_port = environ.get("SERVER_PORT", "")
    req.remote_host = environ.get("REMOTE_HOST", "")
    content_length_str = environ.get("CONTENT_LENGTH", "")
    req.content_length = int(content_length_str) if content_length_str.isdigit() else 0
    req.script_name = environ.get("SCRIPT_NAME", "")
    req.server_protocol = environ.get("SERVER_PROTOCOL", "")
    req.server_software = environ.get("SERVER_SOFTWARE", "")
    req.request_method = environ.get("REQUEST_METHOD", "")
    req.query_string = environ.get("QUERY_STRING", "")
    req.query_string_dict = {
        key: value[0] if len(value) == 1 else value
        for key, value in urllib.parse.parse_qs(req.query_string).items()
    }
    req.remote_addr = environ.get("REMOTE_ADDR", "")
    req.content_type = environ.get("CONTENT_TYPE", "")
    req.http_host = environ.get("HTTP_HOST", "")
    req.http_user_agent = environ.get("HTTP_USER_AGENT", "")
    req.http_accept = environ.get("HTTP_ACCEPT", "")
    req.http_accept_language = environ.get(
        "HTTP_ACCEPT_LANGUAGE", _config.accept_default_lang
    )
    req.http_accept_encoding = environ.get("HTTP_ACCEPT_ENCODING", "")
    req.http_connection = environ.get("HTTP_CONNECTION", "")
    req.http_cookie = environ.get("HTTP_COOKIE", "")

    cookie_dict = {}
    splitted_cookie = req.http_cookie.split(";")
    if splitted_cookie[0] is not "":
        for cookie in splitted_cookie:
            key, value = cookie.split("=")
            cookie_dict[key.strip()] = value

    req.parsed_cookie = cookie_dict

    req.authorization_header = environ.get("HTTP_AUTHORIZATION", None)
    req.http_upgrade_insecure_requests = environ.get(
        "HTTP_UPGRADE_INSECURE_REQUESTS", ""
    )
    req.http_sec_fetch_dest = environ.get("HTTP_SEC_FETCH_DEST", "")
    req.http_sec_fetch_mode = environ.get("HTTP_SEC_FETCH_MODE", "")
    req.http_sec_fetch_site = environ.get("HTTP_SEC_FETCH_SITE", "")
    req.http_sec_fetch_user = environ.get("HTTP_SEC_FETCH_USER", "")
    req.http_priority = environ.get("HTTP_PRIORITY", "")

    try:
        if req.content_length > 0:
            req.body_stream = environ.get("wsgi.input", None)
            if req.body_stream:
                req.body_raw = req.body_stream.read(req.content_length).decode("utf-8")
                req.body_str = str(req.body_raw)                
                req.body_json = json.loads(req.body_raw)
    except json.JSONDecodeError:
        req.body_json = None
    except RecursionError:
        req.body_json = None
    except TypeError:
        req.body_json = None
    except UnicodeDecodeError:
        req.body_raw = None
        req.body_str = None
        req.body_json = None
    except Exception as err:
        print(type(err))
        _logger.log_error(
            "Error creating body from WSGI input.", traceback.format_exc(), 100, req
        )

    return req


def _evaluate_content(resp: Response):
    if resp.json is not None:
        return [resp.json.encode("utf-8")]
    if resp.to_convert is not None:
        return [json.dumps(resp.to_convert).encode("utf-8")]
    if resp.custom_bytearray is not None:
        return [resp.custom_bytearray]
    return []


def set_config(config: PConfig):
    """
    Configer the application.

    Args:
        config (PConfig): Set a new configuration. Only needed if a change is required.
    """
    global _config, _logger
    _config = config
    _logger.log_level = _config.log_level
    _logger.log_info("set_config, config changed", 1)


def set_path_cache(cache: PathCache):
    """
    Define a own path cache implementation.

    Args:
        cache (PathCache): Set a new cache implementation. Only needed if a own implementation is needed.
    """
    global _path_cache
    _path_cache = cache


def set_logger(logger: PLogger):
    """
    Use a own logger.

    Args:
        logger (PLogger): Set a new logger. Only needed if a own logger is required.
    """
    global _logger
    _logger = logger


def set_initialize_hook(hook: PHook):
    """
    Configer the early start of the DAG.

    Args:
        hook (PHook): Set a hook which is processed bevor starting the DAG. You can chain hooks: Hook1 >> Hook2 if you need more than one.
    """
    global _initialize_hook
    _initialize_hook = hook
    _logger.log_info("set_initalize_hook, hook was set.", 2)


def set_finalize_hook(hook: PHook):
    """
    Configer the end of the DAG.

    Args:
        hook (PHook): Set a hook which s processed at the end of the DAG. You can chain hooks: Hook1 >> Hook2 if you need more than one.
    """
    global _finalize_hook
    _finalize_hook = hook
    _logger.log_info("set_finalize_hook, hook was set.", 2)


def list_supported_http_status_codes() -> list:
    """
    Ask for the supported http status codes in Papys.

    Returns:
        list: A list of strings with the supported http status codes.
    """
    return list(_http_code_mapping.keys())


def list_supported_http_methods() -> list:
    """
    Ask for the supported http method codes in Papys.

    Returns:
        list: A list of strings with the supported http methods. Actually: GET, POST, PUT, DELETE.
    """
    return ["GET", "POST", "PUT", "DELETE"]


def add_route(route: PRoute):
    """
    Add a route to the application.

    Args:
        route (PRoute): The route to add to Papys. You can chain routes: Route1 | Route2.

    Raises:
        Exception: If the route can not be added. The exception will also be logged.
    """
    try:
        _logger.log_info("add_route, start adding a route.", 3)
        _add_actions(route.path, route.actions, route.hook)
        _logger.log_info("add_route, _add_actions done.", 4)
        _add_sub_routes(route.path, route.sub_routes, route.hook)
        _logger.log_info("add_route, _add_sub_routes, done.", 5)
    except Exception as err:
        _logger.log_error("Error during adding a route.", traceback.format_exc(), 99)
        raise


def app(environ, start_response):
    """
    The WSGI compatible function which can be given to a WSGI server for the execution.
    See: https://peps.python.org/pep-0333/

    Args:
        environ (dict): Dict with the request information.
        start_response (function): A function which can be called for the response.
    """
    try:
        req = _create_request(environ)
        resp = Response()

        initialize_result = (
            _execute_hook(_initialize_hook, req, resp)
            if _initialize_hook is not None
            else (True, 0, None, None)
        )

        if initialize_result[0]:
            match environ.get("REQUEST_METHOD"):
                case "GET":
                    result = _handle_request(
                        req,
                        resp,
                        _get_actions,
                        [_post_actions, _put_actions, _delete_actions],
                    )
                    resp.status_code = result[0]
                case "POST":
                    result = _handle_request(
                        req,
                        resp,
                        _post_actions,
                        [_get_actions, _put_actions, _delete_actions],
                    )
                    resp.status_code = result[0]
                case "PUT":
                    result = _handle_request(
                        req,
                        resp,
                        _put_actions,
                        [_get_actions, _post_actions, _delete_actions],
                    )
                    resp.status_code = result[0]
                case "DELETE":
                    result = _handle_request(
                        req,
                        resp,
                        _delete_actions,
                        [_get_actions, _post_actions, _put_actions],
                    )
                    resp.status_code = result[0]

            if _finalize_hook:
                finalize_result = _execute_hook(_finalize_hook, req, resp)
                resp.status_code = finalize_result[1]

        else:
            resp.status_code = initialize_result[1]

        resp.set_header("Content-type", resp.content_type)

        match resp.status_code, environ.get("REQUEST_METHOD"), _config.post_convert_201:
            case 200, "POST", True:
                http_status_code = _http_code_mapping.get(
                    201, "500 Internal Server Error"
                )
            case _:
                http_status_code = _http_code_mapping.get(
                    resp.status_code, "500 Internal Server Error"
                )

        start_response(http_status_code, resp.list_headers())

        if 400 <= resp.status_code < 500:
            resp.reset_content()

        match resp.status_code:
            case 500:
                if _config.return_error500_body:
                    return [
                        json.dumps(
                            {
                                "errorCode": resp.status_code,
                                "errorMessage": str(resp.error),
                            }
                        ).encode("utf-8")
                    ]
                else:                    
                    return _evaluate_content(resp)
            case _:                
                return _evaluate_content(resp)

    except Exception as err:
        _logger.log_error(
            "Error during handling a request.", traceback.format_exc(), 101, req
        )
        start_response(_http_code_mapping.get(500, ""), resp.list_headers())
        return []



import traceback
from typing import Tuple
from papys.request_response import Request, Response


class PAction:
    """
    Class providing a process function to execute any kind of actions.
    There are various implementations of this base class for different standard tasks.
    Feel free to make your own implementations. This base implementation gives you the option of passing your action as a function.
    In many cases, it is therefore not necessary to implement your own class.
    It is usually sufficient to implement the action logic in a function and have it executed with this base class.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        pf (function): The function to execute: func(Request, Response) -> Tuple[int, Request, Response]. It's recommended to use valid http status code for the first value in the response tuple. But you are free to use any integer. Handle it careful in the DAG.
    """

    def __init__(self, name: str = "", pf=None):
        self.name = name
        self.process_function = pf
        self._sub_actions = None

    @property
    def sub_actions(self):
        return self._sub_actions

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            return self.process_function(req, resp)
        except Exception as err:
            req.logger.log_error(
                "PAction function could not processed.", traceback.format_exc(), 120, req
            )
            resp.is_error = True
            resp.error = err            
            return 500, req, resp

    # >>
    def __rshift__(self, other):
        # TODO: Check type. List with Tuple or direct Tuple.
        self._sub_actions = other
        return self



import traceback
from typing import Tuple
from papys.actions.core import PAction
from papys.request_response import Request, Response


class PostBounceAction(PAction):
    """
    Class providing a simple post mockup.
    This implementation does nothing other than return a body that has been sent by a client 1:1,
    i.e. write it to the response. The class is purely for test purposes.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        status_codes (dict): You can customize the returned status in case of success and an error. Default: { 'success': 200, 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        status_codes={"success": 200, "error": 500},
    ):
        self._status_codes = status_codes
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            resp.to_convert = req.body_json
            return self._status_codes["success"], req, resp
        except Exception as err:
            req.logger.log_error(
                "PostBounceAction could not processed.", traceback.format_exc(), 140, req
            )
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp


class DummyAction(PAction):
    """
    Class providing a dummy action.
    This implementation does exactly nothing.
    It only returns the parameters with status code 200.
    It is only used for testing purposes.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
    """

    def __init__(self, name: str = ""):
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        return 200, req, resp



import traceback
from typing import Tuple
import requests
from papys.actions.core import PAction
from papys.request_response import Request, Response, COOKIE_SAME_SITE_LAX


class KcOidcAcfCallbackAction(PAction):
    """
    A class that handle the Keycloak callback call after the user login.
    This class is created from KcOIDCFactory. It's recommended to not
    instance this class yourself.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        config (dict): The configuration of the process: {
                "base_url": "",
                "client_id": "",
                "client_secret": "",
                "redirect_url": "",
                "app_redirect_url": "",
            }
        status_codes (dict): You can customize the returned status in case of an error. Default: { 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        config: dict = {},
        status_codes={"error": 500},
    ):
        self._status_codes = status_codes
        self._config = config
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            code = req.query_string_dict.get("code", None)
            if code:
                token_response = requests.post(
                    self._config.get("base_url") + "/protocol/openid-connect/token",
                    data={
                        "grant_type": "authorization_code",
                        "code": code,
                        "redirect_uri": self._config.get("redirect_url"),
                        "client_id": self._config.get("client_id"),
                        "client_secret": self._config.get("client_secret"),
                    },
                )
                token = token_response.json()
                if "access_token" in token:
                    access_token = token["access_token"]
                    refresh_token = token["refresh_token"]

                    resp.add_cookie(
                        "access_token",
                        access_token,
                        resp.create_cookie_attributes(
                            http_only=True,
                            same_site=COOKIE_SAME_SITE_LAX,
                            path="/",
                            secure=True,
                        ),
                    )

                    resp.add_cookie(
                        "refresh_token",
                        refresh_token,
                        resp.create_cookie_attributes(
                            http_only=True,
                            same_site=COOKIE_SAME_SITE_LAX,
                            path="/",
                            secure=True,
                        ),
                    )
                    resp.set_header("Location", self._config.get("app_redirect_url"))
                    return 302, req, resp
                else:
                    req.logger.log_warning(
                        f"Authorization failed, 401. Token: {token}",
                        req=req,
                        log_code=170,
                    )
                    return 401, req, resp
            else:
                req.logger.log_warning("Unauthorized 400", req=req, log_code=171)
                return 400, req, resp

        except Exception as err:
            req.logger.log_error(
                "KcOAuth2CallbackAction could not processed.",
                traceback.format_exc(),
                172,
                req,
            )
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp


class KcOidcAcfLogoutAction(PAction):
    """
    A class handle the logout request from a client.
    This class is created from KcOIDCFactory. It's recommended to not
    instance this class yourself.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        config (dict): The configuration of the action: {
                "base_url": "",
                "client_id": "",
                "client_secret": "",
            }
        status_codes (dict): You can customize the returned status in case of an error. Default: { 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        config: dict = {},
        status_codes={"error": 500},
    ):
        self._status_codes = status_codes
        self._config = config
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            logout_response = requests.post(
                self._config.get("base_url") + "/protocol/openid-connect/logout",
                headers={
                    "Authorization": f"Bearer {req.parsed_cookie.get('access_token', '')}"
                },
                data={
                    "client_id": self._config.get("client_id"),
                    "client_secret": self._config.get("client_secret"),
                    "refresh_token": req.parsed_cookie.get("refresh_token", ""),
                },
            )

            if logout_response.status_code == 204:
                resp.remove_cookie_from_client("access_token")
                resp.remove_cookie_from_client("refresh_token")

                redirect_url = req.query_string_dict.get("redirect_url", None)
                if redirect_url:
                    resp.set_header("Location", redirect_url)
                    return 302, req, resp
                else:
                    resp.to_convert = {"logout": "successful"}
                    return 200, req, resp
            elif logout_response.status_code == 400:
                resp.to_convert = {"logout": "No active user to log out."}
                req.logger.log_warning(
                    "Logout but no activ user.",                    
                    172,
                    req,
                )
                return 400, req, resp
            else:
                req.logger.log_error(
                    "Logout failed",
                    f"IAM status code: {logout_response.status_code}",
                    173,
                    req,
                )
                return 500, req, resp

        except Exception as err:
            req.logger.log_error(
                "KcOIDCLogoutAction could not processed.",
                traceback.format_exc(),
                174,
                req,
            )
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp



import uuid
from datetime import datetime, timezone
from typing import Tuple, Any


class PLogger:
    """
    The default logger of Papys. You can implement you own.
    This logger logs JSON formatted logs to the console using print function.

    Args:
        log_level (int): 1: Errors only. 2: Errors and warnings. 3: All incl. information. Default is 2. You can set this level also in the PConfig.
    """

    def __init__(self, log_level: int = 2):
        self._log_level = log_level

    @property
    def log_level(self) -> int:
        return self._log_level

    @log_level.setter
    def log_level(self, value: int):
        self._log_level = value

    def log_error(
        self, message: str, stack: str = "", log_code: int = 0, req=None
    ) -> str:
        """
        Log as an error.

        Args:
            message (str): The error message.
            stack (str): The error stack as string.
            log_code (int): A number you can refer in your code so you know exactly from where in your code the log entry is come from.
            req (Request): The request provided vom Papys. Giving the request, the Request.process_id is automatically logged. This helps to group log entries to the same client request.

        Returns:
            str: A uuid value which is also logged to the console. Using this id you can track and find the exact log output.
        """

        log_id = str(uuid.uuid4())
        print(
            {                
                "date": datetime.now(timezone.utc).isoformat(),
                "severity": "Error",
                "logId": log_id,
                "logCode": log_code,
                "message": message,
                "stack": stack,
                "processId": req.process_id if req is not None else "",
            }
        )
        return log_id

    def log_warning(self, message: str, log_code: int = 0, req=None) -> str:
        """
        Log as a warning.

        Args:
            message (str): The warning message.
            log_code (int): A number you can refer in your code so you know exactly from where in your code the log entry is come from.
            req (Request): The request provided vom Papys. Giving the request, the Request.process_id is automatically logged. This helps to group log entries to the same client request.

        Returns:
            str: A uuid value which is also logged to the console. Using this id you can track and find the exact log output.
        """

        if self.log_level > 1:
            log_id = str(uuid.uuid4())
            print(
                {                    
                    "date": datetime.now(timezone.utc).isoformat(),
                    "severity": "Warning",
                    "logId": log_id,
                    "logCode": log_code,
                    "message": message,
                    "processId": req.process_id if req is not None else "",
                }
            )
            return log_id

    def log_info(self, message: str, log_code: int = 0, req=None) -> str:
        """
        Log as an information.

        Args:
            message (str): The information message.
            log_code (int): A number you can refer in your code so you know exactly from where in your code the log entry is come from.
            req (Request): The request provided vom Papys. Giving the request, the Request.process_id is automatically logged. This helps to group log entries to the same client request.

        Returns:
            str: A uuid value which is also logged to the console. Using this id you can track and find the exact log output.
        """

        if self.log_level > 2:
            log_id = str(uuid.uuid4())
            print(
                {
                    "date": datetime.now(timezone.utc).isoformat(),
                    "severity": "Info",
                    "logId": log_id,
                    "logCode": log_code,
                    "message": message,
                    "processId": req.process_id if req is not None else "",
                }
            )
            return log_id


class PathCache:
    """
    A simple cache implementation for caching the requested pathes incl. variables.
    Important:
    The cache caches only the path, meaning for which path which (first) action incl. hooks has to called.
    In cases of many endpoints and tousends of clients requesting the same path regulary there is a potential
    to getting faster if this paths as not to be calculated every time.
    To shield the memory the max. size of the cache is limited. After reaching the limit the cache will be cleared and begins from zero.

    Args:
        max_size (int): The maximum of stored pathes in the cache before the cache will be cleared. Default is 100,000.
    """

    def __init__(self, max_size=100000):
        self.max_size = max_size
        self.get_cache = {}
        self.post_cache = {}
        self.put_cache = {}
        self.delete_cache = {}

    def check(self, method: str, path: str) -> Tuple[bool, Any, Any, Any]:
        try:
            match method:
                case "GET":
                    cache_result = self.get_cache.get(path, False)
                    if cache_result == False:
                        return False, None, None, None
                    else:
                        return (True,) + cache_result
                case "POST":
                    cache_result = self.post_cache.get(path, False)
                    if cache_result == False:
                        return False, None, None, None
                    else:
                        return (True,) + cache_result
                case "PUT":
                    cache_result = self.put_cache.get(path, False)
                    if cache_result == False:
                        return False, None, None, None
                    else:
                        return (True,) + cache_result
                case "DELETE":
                    cache_result = self.delete_cache.get(path, False)
                    if cache_result == False:
                        return False, None, None, None
                    else:
                        return (True,) + cache_result
        except:
            return False, None, None, None

    def add(self, method: str, path: str, action: Tuple[Any, Any, Any]):
        match method:
            case "GET":
                if len(self.get_cache) >= self.max_size:
                    self.get_cache.clear()
                self.get_cache[path] = action
            case "POST":
                if len(self.get_cache) >= self.max_size:
                    self.get_cache.clear()
                self.post_cache[path] = action
            case "PUT":
                if len(self.get_cache) >= self.max_size:
                    self.get_cache.clear()
                self.put_cache[path] = action
            case "DELETE":
                if len(self.get_cache) >= self.max_size:
                    self.get_cache.clear()
                self.delete_cache[path] = action



import traceback
from typing import Tuple, Any
from papys.request_response import Request, Response, COOKIE_SAME_SITE_STRICT


class PHook:
    """
    Base class for a hook. This class has to be implemented by a subclass..
    """

    def __init__(self):
        self._sub_hook = None

    @property
    def sub_hook(self):
        return self._sub_hook

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        raise NotImplementedError("This method must be overridden by a subclass!")

    # >>
    def __rshift__(self, other):
        self._sub_hook = other
        return self


class FunctionHook(PHook):
    """
    PHook implementation for a function execution.

    Args:
        hf (function): The function which should be executed. func(Request, Response) -> Tuple[bool, int, Request, Response].
        status_codes (dict): {'error': 500} is default. You can provide an other code to return in case of an exception.
    """

    def __init__(self, hf=None, status_codes: dict = {"error": 500}):
        self.hook_function = hf
        self._status_codes = status_codes
        super().__init__()

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        if self.hook_function is not None:
            try:
                result = self.hook_function(req, resp)
                return result
            except Exception as err:
                req.logger.log_error(
                    "FunctionHook function could not processed.",
                    traceback.format_exc(),
                    130,
                    req,
                )
                return False, self._status_codes["error"], req, resp
        return False, self._status_codes["error"], req, resp


class ParaMapHook(PHook):
    """
    PHook implementation for a parameter variable mapping.
    This hook inserts new entries to Request.path_variables based on the mapping dict.
    Example of a mapping dict:
    {
        "variable_new": "{doc_id}",
        "symbol_new": "simbol",
    }
    Value in brackets refer to an exiting path variable in Request.path_variables.

    Args:
        map (dict): The mapping logic.
        status_codes (dict): {'success': 200, 'error': 500} is default. You can provide an other code to return in case of success or an exception.
    """

    def __init__(
        self,
        map: dict | None = None,
        status_codes: dict = {"success": 200, "error": 500},
    ):
        self.map = map
        self._status_codes = status_codes
        super().__init__()

    def process_hook(
        self, req: Request, resp: Response
    ) -> Tuple[bool, int, Request, Response]:
        if self.map is not None:
            try:
                for key, value in self.map.items():
                    if value.startswith("{") and value.endswith("}"):
                        variable_name = value[1:-1]
                        req.path_variables[key] = req.path_variables.get(
                            variable_name, "not_found"
                        )
                    else:
                        req.path_variables[key] = value

                result = True, self._status_codes["success"], req, resp
                return result
            except Exception as err:
                req.logger.log_error(
                    "ParaMapHook could not processed.", traceback.format_exc(), 131, req
                )
                resp.is_error = True
                resp.error = err
                return False, self._status_codes["error"], req, resp
        return False, self._status_codes["error"], req, resp



import json
import traceback
from typing import Tuple
from papys.actions.core import PAction
from papys.request_response import Request, Response


class ErrorAction(PAction):
    """
    Class providing a basic handling of errors in the DAG.
    The implementation checks whether Request.is_error is True.
    If so, it creates a response body from Response.error_message.
    If not, a response body is created with the message that the action was called but there is no error.
    So make sure that if this error handler is called in the DAG via a status code,
    that a previous handler has also set Response.is_error to True.

    Args:
        name (str): Die argument is only for you to describe the action in the source code. It's optional.
        status_codes (dict): You can customize the returned status in case of success and error. Default: { 'success': 200, 'error': 500 }.
    """

    def __init__(
        self,
        name: str = "",
        status_codes={"success": 200, "error": 500},
    ):
        self._status_codes = status_codes
        super().__init__(name=name)

    def process(self, req: Request, resp: Response) -> Tuple[int, Request, Response]:
        try:
            if resp.is_error:
                resp.json = json.dumps(
                    {"errorCode": resp.status_code, "errorMessage": str(resp.error)}
                )
                resp.status_code = self._status_codes["error"]
                return self._status_codes["error"], req, resp
            else:
                resp.json = json.dumps(
                    {
                        "errorCode": -1,
                        "errorMessage": "ErrorAction called but no error found. Validate your DAG.",
                    }
                )
                resp.status_code = self._status_codes["success"]
                return self._status_codes["success"], req, resp

        except Exception as err:
            req.logger.log_error("ErrorAction could not processed.", traceback.format_exc(), 125, req)
            resp.is_error = True
            resp.error = err
            return self._status_codes["error"], req, resp



import uuid
import email.utils
from datetime import datetime, timezone, timedelta
from papys.utils import PLogger
from papys.config import PConfig

COOKIE_SAME_SITE_STRICT = "Strict"
COOKIE_SAME_SITE_LAX = "Lax"
COOKIE_SAME_SITE_NONE = "None"


class Request:
    """
    A class that holds all information about a request.
    """

    def __init__(self):
        self._path = ""
        self._body_raw = None
        self._body_json = None
        self._body_str = None
        self._body_stream = None
        self._logger: PLogger = PLogger()
        self._path_variables = {}
        self._process_data = {}
        self._start_time = datetime.now(timezone.utc)
        self._process_id = str(uuid.uuid4())
        self._global_config = None
        self._server_name = ""
        self._gateway_interface = ""
        self._server_port = ""
        self._remote_host = ""
        self._content_length = 0
        self._script_name = ""
        self._server_protocol = ""
        self._server_software = ""
        self._request_method = ""
        self._query_string = ""
        self._query_string_dict = {}
        self._remote_addr = ""
        self._content_type = ""
        self._http_host = ""
        self._http_user_agent = ""
        self._http_accept = ""
        self._http_accept_language = ""
        self._http_accept_encoding = ""
        self._http_connection = ""
        self._http_cookie = ""
        self._parsed_cookie: dict | None = None
        self._http_authorization: str | None = None
        self._http_upgrade_insecure_requests = ""
        self._http_sec_fetch_dest = ""
        self._http_sec_fetch_mode = ""
        self._http_sec_fetch_site = ""
        self._http_sec_fetch_user = ""
        self._http_priority = ""
        self._user_info: dict | None = {}
        self._authentication_method: str | None = None

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str):
        self._path = value

    @property
    def body_raw(self) -> str:
        return self._body_raw

    @body_raw.setter
    def body_raw(self, value: str):
        self._body_raw = value

    @property
    def body_json(self) -> str:
        return self._body_json

    @body_json.setter
    def body_json(self, value: str):
        self._body_json = value

    @property
    def body_str(self) -> str:
        return self._body_str

    @body_str.setter
    def body_str(self, value: str):
        self._body_str = value

    @property
    def body_stream(self):
        return self._body_stream

    @body_stream.setter
    def body_stream(self, value):
        self._body_stream = value

    @property
    def logger(self) -> PLogger:
        return self._logger

    @logger.setter
    def logger(self, value: PLogger):
        self._logger = value

    @property
    def path_variables(self) -> dict:
        return self._path_variables

    @path_variables.setter
    def path_variables(self, value: dict):
        self._path_variables = value

    @property
    def process_data(self) -> dict:
        return self._process_data

    @process_data.setter
    def process_data(self, value: dict):
        self._process_data = value

    @property
    def start_time(self) -> datetime:
        return self._start_time

    @start_time.setter
    def start_time(self, value: datetime):
        self._start_time = value

    @property
    def process_id(self) -> str:
        return self._process_id

    @process_id.setter
    def process_id(self, value: str):
        self._process_id = value

    @property
    def global_config(self) -> PConfig:
        return self._global_config

    @global_config.setter
    def global_config(self, value: PConfig):
        self._global_config = value

    @property
    def server_name(self) -> str:
        return self._server_name

    @server_name.setter
    def server_name(self, value: str):
        self._server_name = value

    @property
    def gateway_interface(self) -> str:
        return self._gateway_interface

    @gateway_interface.setter
    def gateway_interface(self, value: str):
        self._gateway_interface = value

    @property
    def server_port(self) -> str:
        return self._server_port

    @server_port.setter
    def server_port(self, value: str):
        self._server_port = value

    @property
    def remote_host(self) -> str:
        return self._remote_host

    @remote_host.setter
    def remote_host(self, value: str):
        self._remote_host = value

    @property
    def content_length(self) -> int:
        return self._content_length

    @content_length.setter
    def content_length(self, value: int):
        self._content_length = value

    @property
    def script_name(self) -> str:
        return self._script_name

    @script_name.setter
    def script_name(self, value: str):
        self._script_name = value

    @property
    def server_protocol(self) -> str:
        return self._server_protocol

    @server_protocol.setter
    def server_protocol(self, value: str):
        self._server_protocol = value

    @property
    def server_software(self) -> str:
        return self._server_software

    @server_software.setter
    def server_software(self, value: str):
        self._server_software = value

    @property
    def request_method(self) -> str:
        return self._request_method

    @request_method.setter
    def request_method(self, value: str):
        self._request_method = value

    @property
    def query_string(self) -> str:
        return self._query_string

    @query_string.setter
    def query_string(self, value: str):
        self._query_string = value

    @property
    def query_string_dict(self) -> dict:
        return self._query_string_dict

    @query_string_dict.setter
    def query_string_dict(self, value: dict):
        self._query_string_dict = value

    @property
    def remote_addr(self) -> str:
        return self._remote_addr

    @remote_addr.setter
    def remote_addr(self, value: str):
        self._remote_addr = value

    @property
    def content_type(self) -> str:
        return self._content_type

    @content_type.setter
    def content_type(self, value: str):
        self._content_type = value

    @property
    def http_host(self) -> str:
        return self._http_host

    @http_host.setter
    def http_host(self, value: str):
        self._http_host = value

    @property
    def http_user_agent(self) -> str:
        return self._http_user_agent

    @http_user_agent.setter
    def http_user_agent(self, value: str):
        self._http_user_agent = value

    @property
    def http_accept(self) -> str:
        return self._http_accept

    @http_accept.setter
    def http_accept(self, value: str):
        self._http_accept = value

    @property
    def http_accept_language(self) -> str:
        return self._http_accept_language

    @http_accept_language.setter
    def http_accept_language(self, value: str):
        self._http_accept_language = value

    @property
    def http_accept_encoding(self) -> str:
        return self._http_accept_encoding

    @http_accept_encoding.setter
    def http_accept_encoding(self, value: str):
        self._http_accept_encoding = value

    @property
    def http_connection(self) -> str:
        return self._http_connection

    @http_connection.setter
    def http_connection(self, value: str):
        self._http_connection = value

    @property
    def http_cookie(self) -> str:
        return self._http_cookie

    @http_cookie.setter
    def http_cookie(self, value: str):
        self._http_cookie = value

    @property
    def parsed_cookie(self) -> dict:
        return self._parsed_cookie

    @parsed_cookie.setter
    def parsed_cookie(self, value: dict):
        self._parsed_cookie = value

    @property
    def authorization_header(self) -> str:
        return self._http_authorization

    @authorization_header.setter
    def authorization_header(self, value: str):
        self._http_authorization = value

    @property
    def http_upgrade_insecure_requests(self) -> str:
        return self._http_upgrade_insecure_requests

    @http_upgrade_insecure_requests.setter
    def http_upgrade_insecure_requests(self, value: str):
        self._http_upgrade_insecure_requests = value

    @property
    def http_sec_fetch_dest(self) -> str:
        return self._http_sec_fetch_dest

    @http_sec_fetch_dest.setter
    def http_sec_fetch_dest(self, value: str):
        self._http_sec_fetch_dest = value

    @property
    def http_sec_fetch_mode(self) -> str:
        return self._http_sec_fetch_mode

    @http_sec_fetch_mode.setter
    def http_sec_fetch_mode(self, value: str):
        self._http_sec_fetch_mode = value

    @property
    def http_sec_fetch_site(self) -> str:
        return self._http_sec_fetch_site

    @http_sec_fetch_site.setter
    def http_sec_fetch_site(self, value: str):
        self._http_sec_fetch_site = value

    @property
    def http_sec_fetch_user(self) -> str:
        return self._http_sec_fetch_user

    @http_sec_fetch_user.setter
    def http_sec_fetch_user(self, value: str):
        self._http_sec_fetch_user = value

    @property
    def http_priority(self) -> str:
        return self._http_priority

    @http_priority.setter
    def http_priority(self, value: str):
        self._http_priority = value

    @property
    def user_info(self) -> dict | None:
        return self._user_info

    @user_info.setter
    def user_info(self, value: dict | None):
        self._user_info = value

    @property
    def authentication_method(self) -> str | None:
        return self._authentication_method

    @authentication_method.setter
    def authentication_method(self, value: str | None):
        self._authentication_method = value

    def get_bearer_token(self) -> str:
        if self.authorization_header:
            parts = self.authorization_header.split()
            return parts[1] if len(parts) > 1 else None
        return None


class Response:
    """
    A class for all information of a response.

    """

    def __init__(self):
        self._status_code = 500
        self._content_type = "application/json"
        self._json = None
        self._to_convert = None
        self._custom_bytearray = None
        self._is_error = False
        self._error = None
        self._headers = {}
        self._cookies = {}

    @property
    def json(self) -> str:
        return self._json

    @json.setter
    def json(self, value: str):
        self._json = value

    @property
    def to_convert(self) -> dict:
        return self._to_convert

    @to_convert.setter
    def to_convert(self, value: dict):
        self._to_convert = value

    @property
    def custom_bytearray(self) -> bytearray:
        return self._custom_bytearray

    @custom_bytearray.setter
    def custom_bytearray(self, value: bytearray):
        self._custom_bytearray = value

    @property
    def status_code(self) -> int:
        return self._status_code

    @status_code.setter
    def status_code(self, value: int):
        self._status_code = value

    @property
    def is_error(self) -> bool:
        return self._is_error

    @is_error.setter
    def is_error(self, value: bool):
        self._is_error = value

    @property
    def error(self) -> object:
        return self._error

    @error.setter
    def error(self, value: object):
        self._error = value

    @property
    def content_type(self) -> str:
        return self._content_type

    @content_type.setter
    def content_type(self, value: str):
        self._content_type = value

    def set_header(self, name: str, value: str):
        self._headers[name] = value

    def list_headers(self) -> list:
        items = self._headers.items()
        headers_list = list(items)
        for key, value in self._cookies.items():
            headers_list.append(("Set-Cookie", f"{key}={value}"))
        return headers_list

    def add_cookie(self, name: str, value: str, attributes: list = []):
        attr_str = "; ".join(attributes)
        self._cookies[name] = value + "; " + attr_str if attr_str else value

    def remove_cookie(self, name: str):
        self._cookies.pop(name)

    def list_cookies(self) -> list:
        items = self._cookies.items()
        return list(items)

    def reset_content(self):
        self.json = None
        self.to_convert = None
        self.custom_bytearray = None

    def create_cookie_attributes(
        self,
        *,
        http_domain: str | None = None,
        expires: datetime | None = None,
        http_only: bool = False,
        max_age: int | None = None,
        partitioned: bool = False,
        path: str | None = None,
        secure: bool | None = None,
        same_site: str | None = None,
    ) -> list:
        """
        A function to create cookie attributes.

        """

        attrs = []
        if http_domain:
            attrs.append(f"Domain={http_domain}")
        if expires:
            attrs.append(f"Expires={email.utils.format_datetime(expires)}")
        if http_only:
            attrs.append("HttpOnly")
        if max_age:
            attrs.append(f"Max-Age={max_age}")
        if partitioned:
            attrs.append("Partitioned")
        if path:
            attrs.append(f"Path={path}")
        if secure:
            attrs.append("Secure")
        if same_site:
            attrs.append(f"SameSite={same_site}")

        return attrs

    def remove_cookie_from_client(self, cookie_name: str):
        self.add_cookie(
            cookie_name,
            "",
            self.create_cookie_attributes(expires=datetime.now() - timedelta(days=2)),
        )



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="papys",
    version="0.4.0",
    author="Ronny Fuchs",
    author_email="info@asderix.com",
    description="Python Rest-API as a DAG",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://papys.asderix.com/",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    license="Apache 2.0",
)
